import cgi
import json
import typing as t

import cattr

from .structs import Request, Response, JSONDict
from .errors import IncompleteResponseError


def _make_headers(content_length: int, encoding: str = "utf-8") -> bytes:
    headers_bytes = bytearray()
    headers = {
        "Content-Length": content_length,
        "Content-Type": f"application/vscode-jsonrpc; charset={encoding}",
    }
    for (key, value) in headers.items():
        headers_bytes += f"{key}: {value}\r\n".encode(encoding)
    headers_bytes += b"\r\n"
    return headers_bytes


def _make_request(
    method: str,
    params: JSONDict = None,
    id: int = None,
    *,
    encoding: str = "utf-8",
) -> bytes:
    request = bytearray()

    # Set up the actual JSONRPC content and encode it.
    content: JSONDict = {"jsonrpc": "2.0", "method": method}
    if params is not None:
        content["params"] = params
    if id is not None:
        content["id"] = id
    encoded_content = json.dumps(content).encode(encoding)

    # Write the headers to the request body
    request += _make_headers(
        content_length=len(encoded_content), encoding=encoding
    )

    # Append the content to the request
    request += encoded_content

    return request


def _make_response(
    id: int,
    result: JSONDict = None,
    error: JSONDict = None,
    *,
    encoding: str = "utf-8",
) -> bytes:
    request = bytearray()

    # Set up the actual JSONRPC content and encode it.
    content: JSONDict = {"jsonrpc": "2.0", "id": id}
    if result is not None:
        content["result"] = result
    if error is not None:
        content["error"] = error
    encoded_content = json.dumps(content).encode(encoding)

    # Write the headers to the request body
    request += _make_headers(
        content_length=len(encoded_content), encoding=encoding
    )

    # Append the content to the request
    request += encoded_content

    return request


def _parse_messages(response: bytes) -> t.Iterator[t.Union[Response, Request]]:
    if b"\r\n\r\n" not in response:
        raise IncompleteResponseError("Incomplete headers")

    header_lines, raw_content = response.split(b"\r\n\r\n", 1)

    # Parse the headers.
    headers = {}
    for header_line in header_lines.split(b"\r\n"):
        key, value = header_line.decode("ascii").split(": ", 1)
        headers[key] = value

    # We will now parse the Content-Type and Content-Length headers. Since for
    # version 3.0 of the Language Server Protocol they're the only ones, we can
    # just verify they're there and not keep them around in the Response
    # object.
    assert set(headers.keys()) == {"Content-Type", "Content-Length"}

    # Content-Type and encoding.
    content_type, metadata = cgi.parse_header(headers["Content-Type"])
    assert content_type == "application/vscode-jsonrpc"
    encoding = metadata["charset"]

    # Content-Length
    content_length = int(headers["Content-Length"])

    # We need to verify that the raw_content is long enough, seeing as we might
    # be getting an incomplete request.
    if len(raw_content) < content_length:
        raise IncompleteResponseError(
            "Not enough bytes to fulfill Content-Length requirements."
        )

    # Take only as many bytes as we need. If there's any remaining, they're
    # the next response's.
    raw_content, next_response = (
        raw_content[:content_length],
        raw_content[content_length:],
    )

    def do_it(data: JSONDict) -> t.Union[Response, Request]:
        del data["jsonrpc"]

        try:
            response = cattr.structure(data, Response)
            return response
        except TypeError:
            pass

        try:
            request = cattr.structure(data, Request)
            return request
        except TypeError:
            pass

        raise RuntimeError(f"{data!r} is neither a Request nor a Response!")

    content = json.loads(raw_content.decode(encoding))
    if isinstance(content, list):
        # This is in response to a batch operation.
        yield from map(do_it, content)
    else:
        yield do_it(content)

    if next_response:
        yield from _parse_messages(next_response)



import enum
import typing as t

from attr import attrs, attrib

# XXX: Replace the non-commented-out code with what's commented out once nested
# types become a thing in mypy.
# JSONValue = t.Union[None, str, int, t.List['JSONValue'], t.Dict[str, 'JSONValue']]
# JSONDict = t.Dict[str, JSONValue]
JSONDict = t.Dict[str, t.Any]

# XXX: We can't have this be both str and int due to `cattrs` restrictions. How
# can we fix this?
# Id = t.Union[str, int]
Id = int


@attrs
class Request:
    method: str = attrib()
    id: t.Optional[Id] = attrib(default=None)
    params: t.Optional[JSONDict] = attrib(default=None)


@attrs
class Response:
    id: t.Optional[Id] = attrib(default=None)
    result: t.Optional[JSONDict] = attrib(default=None)
    error: t.Optional[JSONDict] = attrib(default=None)


class MessageType(enum.IntEnum):
    ERROR = 1
    WARNING = 2
    INFO = 3
    LOG = 4


@attrs
class MessageActionItem:
    title: str = attrib()


@attrs
class TextDocumentItem:
    uri: str = attrib()
    languageId: str = attrib()
    version: int = attrib()
    text: str = attrib()


@attrs
class TextDocumentIdentifier:
    uri: str = attrib()


@attrs
class VersionedTextDocumentIdentifier(TextDocumentIdentifier):
    version: t.Optional[int] = attrib(default=None)


@attrs
class Position:
    # NB: These are both zero-based.
    line: int = attrib()
    character: int = attrib()


@attrs
class Range:
    start: Position = attrib()
    end: Position = attrib()

    def __len__(self) -> int:
        raise NotImplementedError(
            "i as a developer am too stupid to figure out how to implement this"
        )


@attrs
class TextDocumentContentChangeEvent:
    text: str = attrib()
    range: t.Optional[Range] = attrib(default=None)
    rangeLength: t.Optional[int] = attrib(default=None)

    def change_range(
        cls, change_start: Position, change_end: Position, change_text: str
    ) -> "TextDocumentContentChangeEvent":
        change_range = Range(change_start, change_end)
        return cls(
            range=change_range, rangeLength=len(change_range), text=change_text
        )

    def change_whole_document(
        cls, change_text: str
    ) -> "TextDocumentContentChangeEvent":
        return cls(text=change_text)


@attrs
class TextDocumentPosition:
    textDocument: TextDocumentIdentifier = attrib()
    position: Position = attrib()


class CompletionTriggerKind(enum.IntEnum):
    INVOKED = 1
    TRIGGER_CHARACTER = 2
    TRIGGER_FOR_INCOMPLETE_COMPLETIONS = 3


@attrs
class CompletionContext:
    triggerKind: CompletionTriggerKind = attrib()
    triggerCharacter: t.Optional[str] = attrib(default=None)


class MarkupKind(enum.Enum):
    PLAINTEXT = "plaintext"
    MARKDOWN = "markdown"


@attrs
class MarkupContent:
    kind: MarkupKind = attrib()
    value: str = attrib()


@attrs
class TextEdit:
    range: Range = attrib()
    newText: str = attrib()


@attrs
class Command:
    title: str = attrib()
    command: str = attrib()
    arguments: t.Optional[t.List[t.Any]] = attrib(default=None)


class InsertTextFormat(enum.IntEnum):
    PLAIN_TEXT = 1
    SNIPPET = 2


@attrs
class CompletionItem:
    label: str = attrib()
    # TODO: implement CompletionItemKind.
    kind: t.Optional[int] = attrib(default=None)
    detail: t.Optional[str] = attrib(default=None)
    # FIXME: Allow `t.Union[str, MarkupContent]` here by defining a cattrs custom
    # loads hook.
    documentation: t.Optional[str] = attrib(default=None)
    deprecated: t.Optional[bool] = attrib(default=None)
    preselect: t.Optional[bool] = attrib(default=None)
    sortText: t.Optional[str] = attrib(default=None)
    filterText: t.Optional[str] = attrib(default=None)
    insertText: t.Optional[str] = attrib(default=None)
    insertTextFormat: t.Optional[InsertTextFormat] = attrib(default=None)
    textEdit: t.Optional[TextEdit] = attrib(default=None)
    additionalTextEdits: t.Optional[t.List[TextEdit]] = attrib(default=None)
    commitCharacters: t.Optional[t.List[str]] = attrib(default=None)
    command: t.Optional[Command] = attrib(default=None)
    data: t.Optional[t.Any] = attrib(default=None)


@attrs
class CompletionList:
    isIncomplete: bool = attrib()
    items: t.List[CompletionItem] = attrib()


class TextDocumentSaveReason(enum.IntEnum):
    MANUAL = 1
    AFTER_DELAY = 2
    FOCUS_OUT = 3



#!/usr/bin/env python3
# just a canvas to play in.
import os
import pprint
import socket
import urllib.request

from .client import Client
from .errors import IncompleteResponseError
from .events import Initialized, Shutdown, ShowMessageRequest, Completion
from .structs import (
    CompletionTriggerKind,
    TextDocumentItem,
    TextDocumentPosition,
    CompletionContext,
    TextDocumentIdentifier,
    VersionedTextDocumentIdentifier,
    TextDocumentContentChangeEvent,
    TextDocumentSaveReason,
    Position,
)


def main() -> None:
    sock = socket.socket()
    sock.connect(("localhost", int(os.environ.get("PORT", 8080))))

    client = Client(trace="verbose")

    file_path = "./playground.py"
    file_uri = "file://" + urllib.request.pathname2url(
        os.path.abspath(file_path)
    )
    print("File URI:", file_uri)

    while True:
        sock.sendall(client.send())

        try:
            data = sock.recv(4096)
            if not data:
                break
            events = list(client.recv(data))
        except IncompleteResponseError as e:
            continue

        for event in events:
            if isinstance(event, Initialized):
                print("Initialized!")

                print("Server capabilities:")
                pprint.pprint(event.capabilities)

                client.did_open(
                    TextDocumentItem(
                        uri=file_uri,
                        languageId="python",
                        text=open(file_path).read(),
                        version=0,
                    )
                )

                client.completions(
                    text_document_position=TextDocumentPosition(
                        textDocument=TextDocumentIdentifier(uri=file_uri),
                        position=Position(
                            line=5, character=4 + len("struct.") + 1
                        ),
                    ),
                    context=CompletionContext(
                        triggerKind=CompletionTriggerKind.INVOKED
                    ),
                )
            elif isinstance(event, Shutdown):
                print("Shutdown and exiting")
                client.exit()
            elif isinstance(event, Completion):
                print("Completions:")
                pprint.pprint(
                    [item.label for item in event.completion_list.items]
                )

                client.did_close(
                    text_document=TextDocumentIdentifier(uri=file_uri)
                )

                client.will_save(
                    text_document=TextDocumentIdentifier(uri=file_uri),
                    reason=TextDocumentSaveReason.MANUAL,
                )

                client.did_save(
                    text_document=TextDocumentIdentifier(uri=file_uri)
                )

                client.shutdown()
                client.cancel_last_request()
            else:
                raise NotImplementedError(event)


if __name__ == "__main__":
    main()
